@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2016-2017 Canonical, Ltd.
+ * Copyright (C) 2016-2021 Canonical, Ltd.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -24,6 +24,14 @@
  */
 #include "stress-ng.h"
 
+static const stress_help_t help[] = {
+	{ NULL,	"cpu-online N",		"start N workers offlining/onlining the CPUs" },
+	{ NULL,	"cpu-online-ops N",	"stop after N offline/online operations" },
+	{ NULL,	NULL,			NULL }
+};
+
+#define STRESS_CPU_ONLINE_MAX_CPUS	(65536)
+
 #if defined(__linux__)
 
 /*
@@ -31,33 +39,59 @@
  *	set a specified CPUs online or offline
  */
 static int stress_cpu_online_set(
-	const args_t *args,
-	const int32_t cpu,
+	const stress_args_t *args,
+	const uint32_t cpu,
 	const int setting)
 {
 	char filename[PATH_MAX];
-	char data[3] = { '0' + setting, '\n', 0 };
+	char data[3] = { '0' + (char)setting, '\n', 0 };
 	ssize_t ret;
 
 	(void)snprintf(filename, sizeof(filename),
-		"/sys/devices/system/cpu/cpu%" PRId32 "/online", cpu);
+		"/sys/devices/system/cpu/cpu%" PRIu32 "/online", cpu);
 
 	ret = system_write(filename, data, sizeof data);
 	if ((ret < 0) &&
-	    ((ret != -EAGAIN) && (ret != -EINTR) && (ret != -EBUSY))) {
-		pr_fail_err("write");
+	    ((ret != -EAGAIN) && (ret != -EINTR) &&
+	     (ret != -EBUSY) && (ret != -EOPNOTSUPP))) {
+		pr_fail("%s: write failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		return EXIT_FAILURE;
 	}
 	return EXIT_SUCCESS;
 }
 
+/*
+ *  stress_cpu_online_supported()
+ *      check if we can run this as root
+ */
+static int stress_cpu_online_supported(const char *name)
+{
+	ssize_t ret;
+
+	if (geteuid() != 0) {
+		pr_inf_skip("%s stressor will be skipped, "
+		       "need to be running as root for this stressor\n", name);
+		return -1;
+	}
+
+	ret = system_write("/sys/devices/system/cpu/cpu1/online", "1\n", 2);
+	if (ret < 0) {
+		pr_inf_skip("%s stressor will be skipped, "
+		       "cannot write to cpu1 online sysfs control file\n", name);
+		return -1;
+	}
+
+	return 0;
+}
+
 /*
  *  stress_cpu_online
  *	stress twiddling CPUs online/offline
  */
-int stress_cpu_online(const args_t *args)
+static int stress_cpu_online(const stress_args_t *args)
 {
-	const int32_t cpus = stress_get_processors_configured();
+	int32_t cpus = stress_get_processors_configured();
 	int32_t i, cpu_online_count = 0;
 	bool *cpu_online;
 	int rc = EXIT_SUCCESS;
@@ -70,13 +104,19 @@ int stress_cpu_online(const args_t *args)
 		return EXIT_SUCCESS;
 	}
 
-	if ((cpus < 1) || (cpus > 65536)) {
-		pr_inf("%s: too few or too many CPUs (found %" PRId32 ")\n",
+	if (cpus < 1) {
+		pr_fail("%s: too few CPUs (detected %" PRId32 ")\n",
 			args->name, cpus);
 		return EXIT_FAILURE;
 	}
+	if (cpus > STRESS_CPU_ONLINE_MAX_CPUS) {
+		pr_inf("%s: more than %" PRId32 " CPUs detected, "
+			"limiting to %d\n",
+			args->name, cpus, STRESS_CPU_ONLINE_MAX_CPUS);
+		cpus = STRESS_CPU_ONLINE_MAX_CPUS;
+	}
 
-	cpu_online = calloc(cpus, sizeof(bool));
+	cpu_online = calloc((size_t)cpus, sizeof(*cpu_online));
 	if (!cpu_online) {
 		pr_err("%s: out of memory\n", args->name);
 		return EXIT_NO_RESOURCE;
@@ -103,12 +143,36 @@ int stress_cpu_online(const args_t *args)
 		free(cpu_online);
 		return EXIT_FAILURE;
 	}
+	if ((args->num_instances > 1) &&
+	    (g_opt_flags & OPT_FLAGS_CPU_ONLINE_ALL)) {
+		if (args->instance == 0) {
+			pr_inf("%s: disabling --cpu-online-all option because "
+			       "more than 1 %s stressor is being invoked\n",
+				args->name, args->name);
+		}
+		g_opt_flags &= ~OPT_FLAGS_CPU_ONLINE_ALL;
+	}
+
+	if ((args->instance == 0) &&
+	    (g_opt_flags & OPT_FLAGS_CPU_ONLINE_ALL)) {
+		pr_inf("%s: exercising all %" PRId32 " cpus\n",
+			args->name, cpu_online_count + 1);
+	}
+
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 
 	/*
 	 *  Now randomly offline/online them all
 	 */
 	do {
-		unsigned long cpu = mwc32() % cpus;
+		uint32_t cpu = stress_mwc32() % (uint32_t)cpus;
+
+		/*
+		 * Only allow CPU 0 to be offlined if OPT_FLAGS_CPU_ONLINE_ALL
+		 * --cpu-online-all has been enabled
+		 */
+		if ((cpu == 0) && !(g_opt_flags & OPT_FLAGS_CPU_ONLINE_ALL))
+			continue;
 		if (cpu_online[cpu]) {
 			rc = stress_cpu_online_set(args, cpu, 0);
 			if (rc != EXIT_SUCCESS)
@@ -118,22 +182,32 @@ int stress_cpu_online(const args_t *args)
 				break;
 			inc_counter(args);
 		}
-	} while (keep_stressing());
+	} while (keep_stressing(args));
+
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 
 	/*
 	 *  Force CPUs all back online
 	 */
 	for (i = 0; i < cpus; i++) {
 		if (cpu_online[i])
-			(void)stress_cpu_online_set(args, i, 1);
+			(void)stress_cpu_online_set(args, (uint32_t)i, 1);
 	}
 	free(cpu_online);
 
 	return rc;
 }
+
+stressor_info_t stress_cpu_online_info = {
+	.stressor = stress_cpu_online,
+	.supported = stress_cpu_online_supported,
+	.class = CLASS_CPU | CLASS_OS | CLASS_PATHOLOGICAL,
+	.help = help
+};
 #else
-int stress_cpu_online(const args_t *args)
-{
-	return stress_not_implemented(args);
-}
+stressor_info_t stress_cpu_online_info = {
+	.stressor = stress_not_implemented,
+	.class = CLASS_CPU | CLASS_OS | CLASS_PATHOLOGICAL,
+	.help = help
+};
 #endif
