@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013-2017 Canonical, Ltd.
+ * Copyright (C) 2013-2021 Canonical, Ltd.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -24,16 +24,73 @@
  */
 #include "stress-ng.h"
 
-typedef void (*stress_iomix_func)(const args_t *args, const int fd, const off_t iomix_bytes);
+typedef void (*stress_iomix_func)(const stress_args_t *args, const int fd, const off_t iomix_bytes);
 
-void stress_set_iomix_bytes(const char *opt)
+static const stress_help_t help[] = {
+	{ NULL,	"iomix N",	 "start N workers that have a mix of I/O operations" },
+	{ NULL,	"iomix-bytes N", "write N bytes per iomix worker (default is 1GB)" },
+	{ NULL,	"iomix-ops N",	 "stop iomix workers after N iomix bogo operations" },
+	{ NULL, NULL,		 NULL }
+};
+
+static int stress_set_iomix_bytes(const char *opt)
 {
 	off_t iomix_bytes;
 
-	iomix_bytes = (off_t)get_uint64_byte_filesystem(opt, 1);
-	check_range_bytes("iomix-bytes", iomix_bytes,
+	iomix_bytes = (off_t)stress_get_uint64_byte_filesystem(opt, 1);
+	stress_check_range_bytes("iomix-bytes", (uint64_t)iomix_bytes,
 		MIN_IOMIX_BYTES, MAX_IOMIX_BYTES);
-	set_setting("iomix-bytes", TYPE_ID_OFF_T, &iomix_bytes);
+	return stress_set_setting("iomix-bytes", TYPE_ID_OFF_T, &iomix_bytes);
+}
+
+/*
+ *  stress_iomix_rnd_offset()
+ *	generate a random offset between 0..max-1
+ */
+static off_t stress_iomix_rnd_offset(const off_t max)
+{
+	return (off_t)(stress_mwc64() % max);
+}
+
+/*
+ *  stress_iomix_fsync_min_1Hz()
+ *	sync written data at most every once a second while
+ *	trying to minimize the number time get calls
+ */
+static void stress_iomix_fsync_min_1Hz(const int fd)
+{
+	static double time_last = -1.0;
+	static int counter = 0;
+	static int counter_max = 1;
+
+	if (time_last <= 0.0)
+		time_last = stress_time_now() + 1.0;
+
+	if (counter++ >= counter_max) {
+		const double now = stress_time_now();
+		const double delta = now - time_last;
+
+		/* Less than 1Hz? try again */
+		if (delta < 1.0)
+			return;
+
+		counter_max = (int)((double)counter / delta);
+
+		counter = 0;
+		time_last = now;
+
+		switch (stress_mwc8() % 3) {
+		case 0:
+			(void)shim_fsync(fd);
+			break;
+		case 1:
+			(void)shim_fdatasync(fd);
+			break;
+		case 2:
+			(void)sync();
+			break;
+		}
+	}
 }
 
 /*
@@ -41,43 +98,53 @@ void stress_set_iomix_bytes(const char *opt)
  *	bursty sequential writes
  */
 static void stress_iomix_wr_seq_bursts(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
 	do {
 		off_t ret, posn;
-		const int n = mwc8();
+		const int n = stress_mwc8();
 		int i;
 		struct timeval tv;
 
-		posn = mwc64() % iomix_bytes;
+		posn = stress_iomix_rnd_offset(iomix_bytes);
 		ret = lseek(fd, posn, SEEK_SET);
 		if (ret < 0) {
-			pr_fail("seek");
+			pr_fail("%s: lseek failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			return;
 		}
+#if defined(HAVE_POSIX_FADVISE) &&      \
+    defined(POSIX_FADV_SEQUENTIAL)
+		if (posn < iomix_bytes)
+			(void)posix_fadvise(fd, posn, iomix_bytes - posn, POSIX_FADV_SEQUENTIAL);
+#endif
 		for (i = 0; (i < n) && (posn < iomix_bytes); i++) {
 			char buffer[512];
 			ssize_t rc;
-			const size_t len = 1 + (mwc32() & (sizeof(buffer) - 1));
+			const size_t len = 1 + (stress_mwc32() & (sizeof(buffer) - 1));
 
 			stress_strnrnd(buffer, len);
 
 			rc = write(fd, buffer, len);
 			if (rc < 0) {
-				pr_fail("write");
-				return;
+				if (errno != EPERM) {
+					pr_fail("%s: write failed, errno=%d (%s)\n",
+						args->name, errno, strerror(errno));
+					return;
+				}
 			}
 			posn += rc;
-			inc_counter(args);
-			if (!keep_stressing())
+			if (!keep_stressing(args))
 				return;
+			inc_counter(args);
+			stress_iomix_fsync_min_1Hz(fd);
 		}
 		tv.tv_sec = 0;
-		tv.tv_usec = mwc32() % 1000000;
+		tv.tv_usec = stress_mwc32() % 1000000;
 		(void)select(0, NULL, NULL, NULL, &tv);
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 }
 
 /*
@@ -85,43 +152,52 @@ static void stress_iomix_wr_seq_bursts(
  *	bursty random writes
  */
 static void stress_iomix_wr_rnd_bursts(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
+#if defined(HAVE_POSIX_FADVISE) &&      \
+    defined(POSIX_FADV_RANDOM)
+	(void)posix_fadvise(fd, 0, iomix_bytes, POSIX_FADV_RANDOM);
+#endif
 	do {
-		const int n = mwc8();
+		const int n = stress_mwc8();
 		int i;
 		struct timeval tv;
 
 		for (i = 0; i < n; i++) {
 			char buffer[512];
 			ssize_t rc;
-			const size_t len = 1 + (mwc32() & (sizeof(buffer) - 1));
+			const size_t len = 1 + (stress_mwc32() & (sizeof(buffer) - 1));
 			off_t ret, posn;
 
-			posn = mwc64() % iomix_bytes;
+			posn = stress_iomix_rnd_offset(iomix_bytes);
 			ret = lseek(fd, posn, SEEK_SET);
 			if (ret < 0) {
-				pr_fail("seek");
+				pr_fail("%s: lseek failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 				return;
 			}
 
 			stress_strnrnd(buffer, len);
 			rc = write(fd, buffer, len);
 			if (rc < 0) {
-				pr_fail("write");
-				return;
+				if (errno != EPERM) {
+					pr_fail("%s: write failed, errno=%d (%s)\n",
+						args->name, errno, strerror(errno));
+					return;
+				}
 			}
-			inc_counter(args);
-			if (!keep_stressing())
+			if (!keep_stressing(args))
 				return;
+			inc_counter(args);
+			stress_iomix_fsync_min_1Hz(fd);
 		}
-		tv.tv_sec = mwc32() % 2;
-		tv.tv_usec = mwc32() % 1000000;
+		tv.tv_sec = stress_mwc32() % 2;
+		tv.tv_usec = stress_mwc32() % 1000000;
 		(void)select(0, NULL, NULL, NULL, &tv);
 
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 }
 
 /*
@@ -129,7 +205,7 @@ static void stress_iomix_wr_rnd_bursts(
  *	slow sequential writes
  */
 static void stress_iomix_wr_seq_slow(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
@@ -138,28 +214,38 @@ static void stress_iomix_wr_seq_slow(
 
 		ret = lseek(fd, 0, SEEK_SET);
 		if (ret < 0) {
-			pr_fail("seek");
+			pr_fail("%s: lseek failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			return;
 		}
+#if defined(HAVE_POSIX_FADVISE) &&      \
+    defined(POSIX_FADV_SEQUENTIAL)
+		if (posn < iomix_bytes)
+			(void)posix_fadvise(fd, posn, iomix_bytes - posn, POSIX_FADV_SEQUENTIAL);
+#endif
 		while (posn < iomix_bytes) {
 			char buffer[512];
 			ssize_t rc;
-			const size_t len = 1 + (mwc32() & (sizeof(buffer) - 1));
+			const size_t len = 1 + (stress_mwc32() & (sizeof(buffer) - 1));
 
 			stress_strnrnd(buffer, len);
 
 			rc = write(fd, buffer, len);
 			if (rc < 0) {
-				pr_fail("write");
-				return;
+				if (errno != EPERM) {
+					pr_fail("%s: write failed, errno=%d (%s)\n",
+						args->name, errno, strerror(errno));
+					return;
+				}
 			}
-			(void)usleep(250000);
+			(void)shim_usleep(250000);
 			posn += rc;
-			inc_counter(args);
-			if (!keep_stressing())
+			if (!keep_stressing(args))
 				return;
+			inc_counter(args);
+			stress_iomix_fsync_min_1Hz(fd);
 		}
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 }
 
 /*
@@ -167,44 +253,49 @@ static void stress_iomix_wr_seq_slow(
  *	bursty sequential reads
  */
 static void stress_iomix_rd_seq_bursts(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
 	do {
 		off_t ret, posn;
-		const int n = mwc8();
+		const int n = stress_mwc8();
 		int i;
 		struct timeval tv;
 
-		posn = mwc64() % iomix_bytes;
+		posn = stress_iomix_rnd_offset(iomix_bytes);
 		ret = lseek(fd, posn, SEEK_SET);
 		if (ret < 0) {
-			pr_fail("seek");
+			pr_fail("%s: lseek failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			return;
 		}
-#if defined(__linux__)
-		(void)posix_fadvise(fd, posn, 1024 * 1024, POSIX_FADV_SEQUENTIAL);
+
+#if defined(HAVE_POSIX_FADVISE) &&      \
+    defined(POSIX_FADV_SEQUENTIAL)
+		if (posn < iomix_bytes)
+			(void)posix_fadvise(fd, posn, iomix_bytes - posn, POSIX_FADV_SEQUENTIAL);
 #endif
 		for (i = 0; (i < n) && (posn < iomix_bytes); i++) {
 			char buffer[512];
 			ssize_t rc;
-			const size_t len = 1 + (mwc32() & (sizeof(buffer) - 1));
+			const size_t len = 1 + (stress_mwc32() & (sizeof(buffer) - 1));
 
 			rc = read(fd, buffer, len);
 			if (rc < 0) {
-				pr_fail("read");
+				pr_fail("%s: read failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 				return;
 			}
 			posn += rc;
-			inc_counter(args);
-			if (!keep_stressing())
+			if (!keep_stressing(args))
 				return;
+			inc_counter(args);
 		}
 		tv.tv_sec = 0;
-		tv.tv_usec = mwc32() % 1000000;
+		tv.tv_usec = stress_mwc32() % 1000000;
 		(void)select(0, NULL, NULL, NULL, &tv);
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 }
 
 /*
@@ -212,44 +303,47 @@ static void stress_iomix_rd_seq_bursts(
  *	bursty random reads
  */
 static void stress_iomix_rd_rnd_bursts(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
 	do {
-		const int n = mwc8();
+		const int n = stress_mwc8();
 		int i;
 		struct timeval tv;
 
 		for (i = 0; i < n; i++) {
 			char buffer[512];
 			ssize_t rc;
-			const size_t len = 1 + (mwc32() & (sizeof(buffer) - 1));
+			const size_t len = 1 + (stress_mwc32() & (sizeof(buffer) - 1));
 			off_t ret, posn;
 
-			posn = mwc64() % iomix_bytes;
-#if defined(__linux__)
-			(void)posix_fadvise(fd, posn, len, POSIX_FADV_RANDOM);
+			posn = stress_iomix_rnd_offset(iomix_bytes);
+#if defined(HAVE_POSIX_FADVISE) &&      \
+    defined(POSIX_FADV_RANDOM)
+			(void)posix_fadvise(fd, posn, (off_t)len, POSIX_FADV_RANDOM);
 #endif
 			ret = lseek(fd, posn, SEEK_SET);
 			if (ret < 0) {
-				pr_fail("seek");
+				pr_fail("%s: lseek failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 				return;
 			}
 
 			rc = read(fd, buffer, len);
 			if (rc < 0) {
-				pr_fail("read");
+				pr_fail("%s: read failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 				return;
 			}
-			inc_counter(args);
-			if (!keep_stressing())
+			if (!keep_stressing(args))
 				return;
+			inc_counter(args);
 		}
-		tv.tv_sec = mwc32() % 3;
-		tv.tv_usec = mwc32() % 1000000;
+		tv.tv_sec = stress_mwc32() % 3;
+		tv.tv_usec = stress_mwc32() % 1000000;
 		(void)select(0, NULL, NULL, NULL, &tv);
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 }
 
 /*
@@ -257,7 +351,7 @@ static void stress_iomix_rd_rnd_bursts(
  *	slow sequential reads
  */
 static void stress_iomix_rd_seq_slow(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
@@ -266,29 +360,34 @@ static void stress_iomix_rd_seq_slow(
 
 		ret = lseek(fd, 0, SEEK_SET);
 		if (ret < 0) {
-			pr_fail("seek");
+			pr_fail("%s: lseek failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			return;
 		}
+#if defined(HAVE_POSIX_FADVISE) &&      \
+    defined(POSIX_FADV_SEQUENTIAL)
+		if (posn < iomix_bytes)
+			(void)posix_fadvise(fd, posn, iomix_bytes - posn, POSIX_FADV_SEQUENTIAL);
+#endif
 		while (posn < iomix_bytes) {
 			char buffer[512];
 			ssize_t rc;
-			const size_t len = 1 + (mwc32() & (sizeof(buffer) - 1));
+			const size_t len = 1 + (stress_mwc32() & (sizeof(buffer) - 1));
 
-#if defined(__linux__)
-			(void)posix_fadvise(fd, posn, len, POSIX_FADV_SEQUENTIAL);
-#endif
 			rc = read(fd, buffer, len);
 			if (rc < 0) {
-				pr_fail("read");
+				pr_fail("%s: read failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 				return;
 			}
-			(void)usleep(333333);
+			(void)shim_usleep(333333);
 			posn += rc;
-			inc_counter(args);
-			if (!keep_stressing())
+			if (!keep_stressing(args))
 				return;
+			inc_counter(args);
+			stress_iomix_fsync_min_1Hz(fd);
 		}
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 }
 
 /*
@@ -296,64 +395,69 @@ static void stress_iomix_rd_seq_slow(
  *	file syncs
  */
 static void stress_iomix_sync(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
 	do {
 		struct timeval tv;
 
-		(void)fsync(fd);
-		inc_counter(args);
-		if (!keep_stressing())
+		(void)shim_fsync(fd);
+		if (!keep_stressing(args))
 			break;
-		tv.tv_sec = mwc32() % 4;
-		tv.tv_usec = mwc32() % 1000000;
+		inc_counter(args);
+		tv.tv_sec = stress_mwc32() % 4;
+		tv.tv_usec = stress_mwc32() % 1000000;
 		(void)select(0, NULL, NULL, NULL, &tv);
-		if (!keep_stressing())
+		if (!keep_stressing(args))
 			break;
 
-#if defined(__linux__)
-		(void)fdatasync(fd);
-		inc_counter(args);
-		if (!keep_stressing())
+#if defined(HAVE_FDATASYNC)
+		(void)shim_fdatasync(fd);
+		/* Exercise illegal fdatasync */
+		(void)shim_fdatasync(-1);
+		if (!keep_stressing(args))
 			break;
-		tv.tv_sec = mwc32() % 4;
-		tv.tv_usec = mwc32() % 1000000;
+		tv.tv_sec = stress_mwc32() % 4;
+		tv.tv_usec = stress_mwc32() % 1000000;
 		(void)select(0, NULL, NULL, NULL, &tv);
-		if (!keep_stressing())
+		if (!keep_stressing(args))
 			break;
 #endif
-#if defined(__linux__)
-		(void)sync_file_range(fd, mwc64() % iomix_bytes, 65536, SYNC_FILE_RANGE_WRITE);
-		inc_counter(args);
-		if (!keep_stressing())
+#if defined(HAVE_SYNC_FILE_RANGE) &&	\
+    defined(SYNC_FILE_RANGE_WRITE)
+		(void)sync_file_range(fd, stress_iomix_rnd_offset(iomix_bytes),
+				65536, SYNC_FILE_RANGE_WRITE);
+		if (!keep_stressing(args))
 			break;
-		tv.tv_sec = mwc32() % 4;
-		tv.tv_usec = mwc32() % 1000000;
+		tv.tv_sec = stress_mwc32() % 4;
+		tv.tv_usec = stress_mwc32() % 1000000;
 		(void)select(0, NULL, NULL, NULL, &tv);
 #else
 		(void)iomix_bytes;
 #endif
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 }
 
-#if defined(__linux__)
+#if defined(HAVE_POSIX_FADVISE) &&	\
+    defined(POSIX_FADV_DONTNEED)
 /*
  *  stress_iomix_bad_advise()
  *	bad fadvise hints
  */
 static void stress_iomix_bad_advise(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
 	do {
-		off_t posn = mwc64() % iomix_bytes;
+		off_t posn = stress_iomix_rnd_offset(iomix_bytes);
 
 		(void)posix_fadvise(fd, posn, 65536, POSIX_FADV_DONTNEED);
-		(void)usleep(100000);
-	} while (keep_stressing());
+		(void)shim_usleep(100000);
+		(void)posix_fadvise(fd, posn, 65536, POSIX_FADV_NORMAL);
+		(void)shim_usleep(100000);
+	} while (keep_stressing(args));
 }
 #endif
 
@@ -362,7 +466,7 @@ static void stress_iomix_bad_advise(
  *	random memory mapped read/writes
  */
 static void stress_iomix_rd_wr_mmap(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
@@ -377,7 +481,8 @@ static void stress_iomix_rd_wr_mmap(
 
 	do {
 		for (i = 0; i < SIZEOF_ARRAY(mmaps); i++) {
-			off_t posn = (mwc64() % iomix_bytes) & ~(page_size - 1);
+			const off_t posn = stress_iomix_rnd_offset(iomix_bytes) & ~((off_t)page_size - 1);
+
 			mmaps[i] = mmap(NULL, page_size,
 					PROT_READ | PROT_WRITE, flags, fd, posn);
 		}
@@ -385,24 +490,24 @@ static void stress_iomix_rd_wr_mmap(
 			if (mmaps[i] != MAP_FAILED) {
 				size_t j;
 				uint64_t sum = 0;
-				uint8_t *buffer =  (uint8_t *)mmaps[i];
+				uint8_t *buffer = (uint8_t *)mmaps[i];
 
 				/* Force page data to be read */
 				for (j = 0; j < page_size; j++)
 					sum += buffer[j];
-				uint64_put(sum);
+				stress_uint64_put(sum);
 
 				stress_strnrnd(mmaps[i], page_size);
 				(void)shim_msync(mmaps[i], page_size,
-					(mwc32() & 1) ? MS_ASYNC : MS_SYNC);
+					stress_mwc1() ? MS_ASYNC : MS_SYNC);
 			}
 		}
-		(void)usleep(100000);
+		(void)shim_usleep(100000);
 		for (i = 0; i < SIZEOF_ARRAY(mmaps); i++) {
 			if (mmaps[i] != MAP_FAILED)
 				(void)munmap(mmaps[i], page_size);
 		}
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 }
 
 /*
@@ -410,7 +515,7 @@ static void stress_iomix_rd_wr_mmap(
  *	lots of small 1 byte writes
  */
 static void stress_iomix_wr_bytes(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
@@ -419,25 +524,68 @@ static void stress_iomix_wr_bytes(
 
 		ret = lseek(fd, 0, SEEK_SET);
 		if (ret < 0) {
-			pr_fail("seek");
+			pr_fail("%s: lseek failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			return;
 		}
 		while (posn < iomix_bytes) {
-			char buffer[1] = { (mwc8() % 26) + 'A' };
+			char buffer[1] = { (stress_mwc8() % 26) + 'A' };
 			ssize_t rc;
 
 			rc = write(fd, buffer, sizeof(buffer));
 			if (rc < 0) {
-				pr_fail("write");
-				return;
-			}
-			(void)usleep(1000);
+				if (errno != EPERM) {
+					pr_fail("%s: write failed, errno=%d (%s)\n",
+						args->name, errno, strerror(errno));
+					return;
+			}	}
+			(void)shim_usleep(1000);
 			posn += rc;
-			inc_counter(args);
-			if (!keep_stressing())
+			if (!keep_stressing(args))
 				return;
+			inc_counter(args);
+			stress_iomix_fsync_min_1Hz(fd);
 		}
-	} while (keep_stressing());
+	} while (keep_stressing(args));
+}
+
+/*
+ *  stress_iomix_wr_rev_bytes()
+ *	lots of small 1 byte writes in reverse order
+ */
+static void stress_iomix_wr_rev_bytes(
+	const stress_args_t *args,
+	const int fd,
+	const off_t iomix_bytes)
+{
+	do {
+		off_t ret, posn = iomix_bytes;
+
+		ret = lseek(fd, 0, SEEK_SET);
+		if (ret < 0) {
+			pr_fail("%s: lseek failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
+			return;
+		}
+		while (posn != 0) {
+			char buffer[1] = { (stress_mwc8() % 26) + 'A' };
+			ssize_t rc;
+
+			rc = write(fd, buffer, sizeof(buffer));
+			if (rc < 0) {
+				if (errno != EPERM) {
+					pr_fail("%s: write failed, errno=%d (%s)\n",
+						args->name, errno, strerror(errno));
+					return;
+			}	}
+			(void)shim_usleep(1000);
+			posn--;
+			if (!keep_stressing(args))
+				return;
+			inc_counter(args);
+			stress_iomix_fsync_min_1Hz(fd);
+		}
+	} while (keep_stressing(args));
 }
 
 /*
@@ -445,7 +593,7 @@ static void stress_iomix_wr_bytes(
  *	lots of small 1 byte reads
  */
 static void stress_iomix_rd_bytes(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
@@ -458,38 +606,61 @@ static void stress_iomix_rd_bytes(
 
 			ret = lseek(fd, posn, SEEK_SET);
 			if (ret < 0) {
-				pr_fail("seek");
+				pr_fail("%s: lseek failed, errno=%d (%s)\n",
+					args->name, errno, strerror(errno));
 				return;
 			}
 
 			rc = read(fd, buffer, sizeof(buffer));
 			if (rc < 0) {
-				pr_fail("write");
-				return;
+				if (errno != EPERM) {
+					pr_fail("%s: write failed, errno=%d (%s)\n",
+						args->name, errno, strerror(errno));
+					return;
+				}
 			}
-			(void)usleep(1000);
+			(void)shim_usleep(1000);
 			posn--;
-			inc_counter(args);
-			if (!keep_stressing())
+			if (!keep_stressing(args))
 				return;
+			inc_counter(args);
 		}
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 }
 
 #if defined(__linux__)
 
+#if defined(FS_APPEND_FL)	|| \
+    defined(FS_COMPR_FL)	|| \
+    defined(FS_IMMUTABLE_FL)	|| \
+    defined(FS_JOURNAL_DATA_FL)	|| \
+    defined(FS_NOATIME_FL)	|| \
+    defined(FS_NOCOW_FL)	|| \
+    defined(FS_NODUMP_FL)	|| \
+    defined(FS_NOTAIL_FL)	|| \
+    defined(FS_SECRM_FL)	|| \
+    defined(FS_SYNC_FL)		|| \
+    defined(FS_UNRM_FL)
 /*
  *  stress_iomix_inode_ioctl()
  *	attempt to set and unset a file based inode flag
  */
-static void stress_iomix_inode_ioctl(const int fd, const int flag, bool *ok)
+static void stress_iomix_inode_ioctl(
+	const stress_args_t *args,
+	const int fd,
+	const int flag,
+	bool *ok)
 {
 	int ret, attr;
 
+	if (!keep_stressing(args))
+		return;
+
+#if defined(FS_IOC_GETFLAGS)
 	ret = ioctl(fd, FS_IOC_GETFLAGS, &attr);
 	if (ret < 0)
 		return;
-
+#if defined(FS_IOC_SETFLAGS)
 	attr |= flag;
 	ret = ioctl(fd, FS_IOC_SETFLAGS, &attr);
 	if (ret < 0)
@@ -499,59 +670,64 @@ static void stress_iomix_inode_ioctl(const int fd, const int flag, bool *ok)
 	ret = ioctl(fd, FS_IOC_SETFLAGS, &attr);
 	if (ret < 0)
 		return;
+#endif
+#endif
 	*ok = true;
 }
+#endif
 
 /*
  *  stress_iomix_inode_flags()
  *	twiddle various inode flags
  */
 static void stress_iomix_inode_flags(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
 	(void)args;
+	(void)fd;
 	(void)iomix_bytes;
 
 	do {
 		bool ok = false;
 #if defined(FS_APPEND_FL)
-		stress_iomix_inode_ioctl(fd, FS_APPEND_FL, &ok);
+		stress_iomix_inode_ioctl(args, fd, FS_APPEND_FL, &ok);
 #endif
 #if defined(FS_COMPR_FL)
-		stress_iomix_inode_ioctl(fd, FS_COMPR_FL, &ok);
+		stress_iomix_inode_ioctl(args, fd, FS_COMPR_FL, &ok);
 #endif
 #if defined(FS_IMMUTABLE_FL)
-		stress_iomix_inode_ioctl(fd, FS_IMMUTABLE_FL, &ok);
+		stress_iomix_inode_ioctl(args, fd, FS_IMMUTABLE_FL, &ok);
 #endif
 #if defined(FS_JOURNAL_DATA_FL)
-		stress_iomix_inode_ioctl(fd, FS_JOURNAL_DATA_FL, &ok);
+		stress_iomix_inode_ioctl(args, fd, FS_JOURNAL_DATA_FL, &ok);
 #endif
 #if defined(FS_NOATIME_FL)
-		stress_iomix_inode_ioctl(fd, FS_NOATIME_FL, &ok);
+		stress_iomix_inode_ioctl(args, fd, FS_NOATIME_FL, &ok);
 #endif
 #if defined(FS_NOCOW_FL)
-		stress_iomix_inode_ioctl(fd, FS_NOCOW_FL, &ok);
+		stress_iomix_inode_ioctl(args, fd, FS_NOCOW_FL, &ok);
 #endif
 #if defined(FS_NODUMP_FL)
-		stress_iomix_inode_ioctl(fd, FS_NODUMP_FL, &ok);
+		stress_iomix_inode_ioctl(args, fd, FS_NODUMP_FL, &ok);
 #endif
 #if defined(FS_NOTAIL_FL)
-		stress_iomix_inode_ioctl(fd, FS_NOTAIL_FL, &ok);
+		stress_iomix_inode_ioctl(args, fd, FS_NOTAIL_FL, &ok);
 #endif
 #if defined(FS_SECRM_FL)
-		stress_iomix_inode_ioctl(fd, FS_SECRM_FL, &ok);
+		stress_iomix_inode_ioctl(args, fd, FS_SECRM_FL, &ok);
 #endif
 #if defined(FS_SYNC_FL)
-		stress_iomix_inode_ioctl(fd, FS_SYNC_FL, &ok);
+		stress_iomix_inode_ioctl(args, fd, FS_SYNC_FL, &ok);
 #endif
 #if defined(FS_UNRM_FL)
-		stress_iomix_inode_ioctl(fd, FS_UNRM_FL, &ok);
+		stress_iomix_inode_ioctl(args, fd, FS_UNRM_FL, &ok);
 #endif
 		if (!ok)
-			exit(EXIT_SUCCESS);
-	} while (keep_stressing());
+			_exit(EXIT_SUCCESS);
+		stress_iomix_fsync_min_1Hz(fd);
+	} while (keep_stressing(args));
 }
 #endif
 
@@ -561,7 +737,7 @@ static void stress_iomix_inode_flags(
  *	occasional file cache dropping
  */
 static void stress_iomix_drop_caches(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd,
 	const off_t iomix_bytes)
 {
@@ -569,23 +745,23 @@ static void stress_iomix_drop_caches(
 	(void)iomix_bytes;
 
 	do {
-		sync();
+		(void)sync();
 		if (system_write("/proc/sys/vm/drop_caches", "1", 1) < 0)
 			(void)pause();
 		(void)sleep(5);
-		if (!keep_stressing())
+		if (!keep_stressing(args))
 			return;
-		sync();
+		(void)sync();
 		if (system_write("/proc/sys/vm/drop_caches", "2", 1) < 0)
 			(void)pause();
 		(void)sleep(5);
-		if (!keep_stressing())
+		if (!keep_stressing(args))
 			return;
-		sync();
+		(void)sync();
 		if (system_write("/proc/sys/vm/drop_caches", "3", 1) < 0)
 			(void)pause();
 		(void)sleep(5);
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 }
 #endif
 
@@ -599,11 +775,12 @@ static stress_iomix_func iomix_funcs[] = {
 	stress_iomix_rd_seq_slow,
 	stress_iomix_rd_seq_slow,
 	stress_iomix_sync,
-#if defined(__linux__)
+#if defined(HAVE_POSIX_FADVISE)
 	stress_iomix_bad_advise,
 #endif
 	stress_iomix_rd_wr_mmap,
 	stress_iomix_wr_bytes,
+	stress_iomix_wr_rev_bytes,
 	stress_iomix_rd_bytes,
 #if defined(__linux__)
 	stress_iomix_inode_flags,
@@ -617,7 +794,7 @@ static stress_iomix_func iomix_funcs[] = {
  *  stress_iomix
  *	stress I/O via random mix of io ops
  */
-int stress_iomix(const args_t *args)
+static int stress_iomix(const stress_args_t *args)
 {
 	int fd, ret;
 	char filename[PATH_MAX];
@@ -632,21 +809,22 @@ int stress_iomix(const args_t *args)
 	counters = (void *)mmap(NULL, sz, PROT_READ | PROT_WRITE,
 			MAP_SHARED | MAP_ANONYMOUS, -1, 0);
 	if (counters == MAP_FAILED) {
-		pr_fail_dbg("mmap");
+		pr_fail("%s: mmap failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		return EXIT_NO_RESOURCE;
 	}
 
-	if (!get_setting("iomix-bytes", &iomix_bytes)) {
+	if (!stress_get_setting("iomix-bytes", &iomix_bytes)) {
 		if (g_opt_flags & OPT_FLAGS_MAXIMIZE)
-			iomix_bytes = MAX_FALLOCATE_BYTES;
+			iomix_bytes = MAXIMIZED_FILE_SIZE;
 		if (g_opt_flags & OPT_FLAGS_MINIMIZE)
-			iomix_bytes = MIN_FALLOCATE_BYTES;
+			iomix_bytes = MIN_IOMIX_BYTES;
 	}
 	iomix_bytes /= args->num_instances;
 	if (iomix_bytes < (off_t)MIN_IOMIX_BYTES)
 		iomix_bytes = (off_t)MIN_IOMIX_BYTES;
 	if (iomix_bytes < (off_t)page_size)
-		iomix_bytes = page_size;
+		iomix_bytes = (off_t)page_size;
 
 	ret = stress_temp_dir_mk_args(args);
 	if (ret < 0) {
@@ -655,11 +833,11 @@ int stress_iomix(const args_t *args)
 	}
 
 	(void)stress_temp_filename_args(args,
-		filename, sizeof(filename), mwc32());
-	(void)umask(0077);
+		filename, sizeof(filename), stress_mwc32());
 	if ((fd = open(filename, O_CREAT | O_RDWR | O_SYNC, S_IRUSR | S_IWUSR)) < 0) {
 		ret = exit_status(errno);
-		pr_fail_err("open");
+		pr_fail("%s: open %s failed, errno=%d (%s)\n",
+			args->name, filename, errno, strerror(errno));
 		goto unmap;
 	}
 	(void)unlink(filename);
@@ -674,7 +852,8 @@ int stress_iomix(const args_t *args)
 			ret = EXIT_NO_RESOURCE;
 		} else {
 			ret = EXIT_FAILURE;
-			pr_fail_err("fallocate");
+			pr_fail("%s: fallocate failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 		}
 		goto tidy;
 	}
@@ -682,53 +861,74 @@ int stress_iomix(const args_t *args)
 	(void)memset(pids, 0, sizeof(pids));
 	(void)memset(counters, 0, sz);
 
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
+
 	for (i = 0; i < SIZEOF_ARRAY(iomix_funcs); i++) {
-		const args_t tmp_args = {
-			.counter = &counters[i],
-			.name = args->name,
-			.max_ops = args->max_ops,
-			.instance = args->instance,
-			.num_instances = args->num_instances,
-			.pid = args->pid,
-			.ppid = args->ppid,
-			.page_size = args->page_size
-		};
+		stress_args_t tmp_args = *args;
+
+		tmp_args.counter = &counters[i];
 
 		pids[i] = fork();
 		if (pids[i] < 0) {
 			goto reap;
 		} else if (pids[i] == 0) {
 			/* Child */
+			(void)sched_settings_apply(true);
 			iomix_funcs[i](&tmp_args, fd, iomix_bytes);
-			(void)kill(args->pid, SIGALRM);
 			_exit(EXIT_SUCCESS);
 		}
 	}
 
 	do {
-		*args->counter = 0;
-		(void)pause();
+		uint64_t c = 0;
+		(void)shim_usleep(5000);
 		for (i = 0; i < SIZEOF_ARRAY(iomix_funcs); i++) {
-			*args->counter += counters[i];
+			c += counters[i];
+			if (UNLIKELY(args->max_ops && c >= args->max_ops)) {
+				set_counter(args, c);
+				goto reap;
+			}
 		}
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 
 	ret = EXIT_SUCCESS;
 reap:
+	set_counter(args, 0);
+	for (i = 0; i < SIZEOF_ARRAY(iomix_funcs); i++) {
+		add_counter(args, counters[i]);
+
+		if (pids[i]) {
+			(void)kill(pids[i], SIGALRM);
+			(void)kill(pids[i], SIGKILL);
+		}
+	}
 	for (i = 0; i < SIZEOF_ARRAY(iomix_funcs); i++) {
 		if (pids[i]) {
 			int status;
 
-			(void)kill(pids[i], SIGKILL);
-			(void)waitpid(pids[i], &status, 0);
+			(void)shim_waitpid(pids[i], &status, 0);
 		}
 	}
 
 tidy:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 	(void)close(fd);
 	(void)stress_temp_dir_rm_args(args);
 unmap:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 	(void)munmap((void *)counters, sz);
 
 	return ret;
 }
+
+static const stress_opt_set_func_t opt_set_funcs[] = {
+	{ OPT_iomix_bytes,	stress_set_iomix_bytes },
+	{ 0,			NULL }
+};
+
+stressor_info_t stress_iomix_info = {
+	.stressor = stress_iomix,
+	.class = CLASS_FILESYSTEM | CLASS_OS,
+	.opt_set_funcs = opt_set_funcs,
+	.help = help
+};
