@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013-2017 Canonical, Ltd.
+ * Copyright (C) 2013-2021 Canonical, Ltd.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -24,35 +24,44 @@
  */
 #include "stress-ng.h"
 
-#if defined(F_OFD_GETLK) && defined(F_OFD_SETLK) && defined(F_OFD_SETLKW) && \
-    defined(F_WRLCK) && defined(F_UNLCK)
+static const stress_help_t help[] = {
+	{ NULL, "lockofd N",	 "start N workers using open file description locking" },
+	{ NULL, "lockofd-ops N", "stop after N lockofd bogo operations" },
+	{ NULL, NULL,		 NULL }
+};
+
+#if defined(F_OFD_GETLK) &&	\
+    defined(F_OFD_SETLK) &&	\
+    defined(F_OFD_SETLKW) && 	\
+    defined(F_WRLCK) &&		\
+    defined(F_UNLCK)
 
 #define LOCK_FILE_SIZE	(1024 * 1024)
 #define LOCK_SIZE	(8)
 #define LOCK_MAX	(1024)
 
-typedef struct lockofd_info {
+typedef struct stress_lockofd_info {
 	off_t	offset;
 	off_t	len;
-	struct lockofd_info *next;
-} lockofd_info_t;
+	struct stress_lockofd_info *next;
+} stress_lockofd_info_t;
 
 typedef struct {
-	lockofd_info_t *head;		/* Head of lockofd_info procs list */
-	lockofd_info_t *tail;		/* Tail of lockofd_info procs list */
-	lockofd_info_t *free;		/* List of free'd lockofd_infos */
+	stress_lockofd_info_t *head;	/* Head of lockofd_info procs list */
+	stress_lockofd_info_t *tail;	/* Tail of lockofd_info procs list */
+	stress_lockofd_info_t *free;	/* List of free'd lockofd_infos */
 	uint64_t length;		/* Length of list */
-} lockofd_info_list_t;
+} stress_lockofd_info_list_t;
 
-static lockofd_info_list_t lockofd_infos;
+static stress_lockofd_info_list_t lockofd_infos;
 
 /*
  *  stress_lockofd_info_new()
  *	allocate a new lockofd_info, add to end of list
  */
-static lockofd_info_t *stress_lockofd_info_new(void)
+static stress_lockofd_info_t *stress_lockofd_info_new(void)
 {
-	lockofd_info_t *new;
+	stress_lockofd_info_t *new;
 
 	if (lockofd_infos.free) {
 		/* Pop an old one off the free list */
@@ -84,7 +93,7 @@ static lockofd_info_t *stress_lockofd_info_new(void)
 static void stress_lockofd_info_head_remove(void)
 {
 	if (lockofd_infos.head) {
-		lockofd_info_t *head = lockofd_infos.head;
+		stress_lockofd_info_t *head = lockofd_infos.head;
 
 		if (lockofd_infos.tail == lockofd_infos.head) {
 			lockofd_infos.tail = NULL;
@@ -108,14 +117,14 @@ static void stress_lockofd_info_head_remove(void)
 static void stress_lockofd_info_free(void)
 {
 	while (lockofd_infos.head) {
-		lockofd_info_t *next = lockofd_infos.head->next;
+		stress_lockofd_info_t *next = lockofd_infos.head->next;
 
 		free(lockofd_infos.head);
 		lockofd_infos.head = next;
 	}
 
 	while (lockofd_infos.free) {
-		lockofd_info_t *next = lockofd_infos.free->next;
+		stress_lockofd_info_t *next = lockofd_infos.free->next;
 
 		free(lockofd_infos.free);
 		lockofd_infos.free = next;
@@ -126,7 +135,7 @@ static void stress_lockofd_info_free(void)
  *  stress_lockofd_unlock()
  *	pop oldest lock record off list and unlock it
  */
-static int stress_lockofd_unlock(const args_t *args, const int fd)
+static int stress_lockofd_unlock(const stress_args_t *args, const int fd)
 {
 	struct flock f;
 
@@ -143,7 +152,8 @@ static int stress_lockofd_unlock(const args_t *args, const int fd)
 	stress_lockofd_info_head_remove();
 
 	if (fcntl(fd, F_OFD_SETLK, &f) < 0) {
-		pr_fail_err("F_SETLK");
+		pr_fail("%s: fcntl F_OFD_SETLK failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		return -1;
 	}
 	return 0;
@@ -154,24 +164,24 @@ static int stress_lockofd_unlock(const args_t *args, const int fd)
  *	hammer advisory lock/unlock to create some file lock contention
  */
 static int stress_lockofd_contention(
-	const args_t *args,
+	const stress_args_t *args,
 	const int fd)
 {
-	mwc_reseed();
+	stress_mwc_reseed();
 
 	do {
 		off_t offset;
 		off_t len;
 		int rc;
-		lockofd_info_t *lockofd_info;
+		stress_lockofd_info_t *lockofd_info;
 		struct flock f;
 
 		if (lockofd_infos.length >= LOCK_MAX)
 			if (stress_lockofd_unlock(args, fd) < 0)
 				return -1;
 
-		len  = (mwc16() + 1) & 0xfff;
-		offset = mwc64() % (LOCK_FILE_SIZE - len);
+		len = (stress_mwc16() + 1) & 0xfff;
+		offset = (off_t)stress_mwc64() % (LOCK_FILE_SIZE - len);
 
 		f.l_type = F_WRLCK;
 		f.l_whence = SEEK_SET;
@@ -187,14 +197,14 @@ static int stress_lockofd_contention(
 
 		lockofd_info = stress_lockofd_info_new();
 		if (!lockofd_info) {
-			pr_fail_err("calloc");
+			pr_fail("%s: calloc failed, out of memory\n", args->name);
 			return -1;
 		}
 		lockofd_info->offset = offset;
 		lockofd_info->len = len;
 
 		inc_counter(args);
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 
 	return 0;
 }
@@ -203,12 +213,12 @@ static int stress_lockofd_contention(
  *  stress_lockofd
  *	stress file locking via advisory locking
  */
-int stress_lockofd(const args_t *args)
+static int stress_lockofd(const stress_args_t *args)
 {
 	int fd, ret = EXIT_FAILURE;
 	pid_t cpid = -1;
 	char filename[PATH_MAX];
-	char dirname[PATH_MAX];
+	char pathname[PATH_MAX];
 	char buffer[4096];
 	off_t offset;
 	ssize_t rc;
@@ -219,11 +229,12 @@ int stress_lockofd(const args_t *args)
 	 *  There will be a race to create the directory
 	 *  so EEXIST is expected on all but one instance
 	 */
-	(void)stress_temp_dir_args(args, dirname, sizeof(dirname));
-	if (mkdir(dirname, S_IRWXU) < 0) {
+	(void)stress_temp_dir_args(args, pathname, sizeof(pathname));
+	if (mkdir(pathname, S_IRWXU) < 0) {
 		if (errno != EEXIST) {
 			ret = exit_status(errno);
-			pr_fail_err("mkdir");
+			pr_fail("%s: mkdir %s failed, errno=%d (%s)\n",
+				args->name, pathname, errno, strerror(errno));
 			return ret;
 		}
 	}
@@ -234,22 +245,24 @@ int stress_lockofd(const args_t *args)
 	 *  stress flock processes
 	 */
 	(void)stress_temp_filename_args(args,
-		filename, sizeof(filename), mwc32());
+		filename, sizeof(filename), stress_mwc32());
 
 	if ((fd = open(filename, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR)) < 0) {
 		ret = exit_status(errno);
-		pr_fail_err("open");
-		(void)rmdir(dirname);
+		pr_fail("%s: open %s failed, errno=%d (%s)\n",
+			args->name, pathname, errno, strerror(errno));
+		(void)rmdir(pathname);
 		return ret;
 	}
 
 	if (lseek(fd, 0, SEEK_SET) < 0) {
-		pr_fail_err("lseek");
+		pr_fail("%s: lseek failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		goto tidy;
 	}
 	for (offset = 0; offset < LOCK_FILE_SIZE; offset += sizeof(buffer)) {
 redo:
-		if (!g_keep_stressing_flag) {
+		if (!keep_stressing_flag()) {
 			ret = EXIT_SUCCESS;
 			goto tidy;
 		}
@@ -258,54 +271,65 @@ redo:
 			if ((errno == EAGAIN) || (errno == EINTR))
 				goto redo;
 			ret = exit_status(errno);
-			pr_fail_err("write");
+			pr_fail("%s: write failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			goto tidy;
 		}
 	}
 
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 again:
 	cpid = fork();
 	if (cpid < 0) {
-		if (!g_keep_stressing_flag) {
-			ret = EXIT_SUCCESS;
-			goto tidy;
-		}
-		if (errno == EAGAIN)
+		if (stress_redo_fork(errno))
 			goto again;
-		pr_fail_err("fork");
+		if (!keep_stressing(args))
+			goto tidy;
+		pr_fail("%s: fork failed, errno=%d (%s)\n",
+			args->name, errno, strerror(errno));
 		goto tidy;
 	}
 	if (cpid == 0) {
 		(void)setpgid(0, g_pgrp);
 		stress_parent_died_alarm();
+		(void)sched_settings_apply(true);
 
 		if (stress_lockofd_contention(args, fd) < 0)
-			exit(EXIT_FAILURE);
+			_exit(EXIT_FAILURE);
 		stress_lockofd_info_free();
-		exit(EXIT_SUCCESS);
+		_exit(EXIT_SUCCESS);
 	}
 	(void)setpgid(cpid, g_pgrp);
 
 	if (stress_lockofd_contention(args, fd) == 0)
 		ret = EXIT_SUCCESS;
 tidy:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
+
 	if (cpid > 0) {
 		int status;
 
 		(void)kill(cpid, SIGKILL);
-		(void)waitpid(cpid, &status, 0);
+		(void)shim_waitpid(cpid, &status, 0);
 	}
 	stress_lockofd_info_free();
 
 	(void)close(fd);
 	(void)unlink(filename);
-	(void)rmdir(dirname);
+	(void)rmdir(pathname);
 
 	return ret;
 }
+
+stressor_info_t stress_lockofd_info = {
+	.stressor = stress_lockofd,
+	.class = CLASS_FILESYSTEM | CLASS_OS,
+	.help = help
+};
 #else
-int stress_lockofd(const args_t *args)
-{
-	return stress_not_implemented(args);
-}
+stressor_info_t stress_lockofd_info = {
+	.stressor = stress_not_implemented,
+	.class = CLASS_FILESYSTEM | CLASS_OS,
+	.help = help
+};
 #endif
