@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013-2017 Canonical, Ltd.
+ * Copyright (C) 2013-2021 Canonical, Ltd.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -24,69 +24,69 @@
  */
 #include "stress-ng.h"
 
-#if defined(HAVE_LIB_PTHREAD) && defined(__linux__)
-
-#include <sys/select.h>
+#if defined(HAVE_LIB_PTHREAD)
 
 typedef struct {
-	const args_t *args;
+	const stress_args_t *args;
 	uint64_t counter;
+	uint64_t sleep_max;
 	pthread_t pthread;
-} ctxt_t;
+} stress_ctxt_t;
 
-static bool thread_terminate;
+static volatile bool thread_terminate;
 static sigset_t set;
 #endif
 
-void stress_set_sleep_max(const char *opt)
+static const stress_help_t help[] = {
+	{ NULL,	"sleep N",	"start N workers performing various duration sleeps" },
+	{ NULL,	"sleep-ops N",	"stop after N bogo sleep operations" },
+	{ NULL,	"sleep-max P",	"create P threads at a time by each worker" },
+	{ NULL,	NULL,		NULL }
+};
+
+static int stress_set_sleep_max(const char *opt)
 {
 	uint64_t sleep_max;
 
-	sleep_max = get_uint64(opt);
-	check_range("sleep-max", sleep_max,
+	sleep_max = stress_get_uint64(opt);
+	stress_check_range("sleep-max", sleep_max,
 		MIN_SLEEP, MAX_SLEEP);
-	set_setting("sleep-max", TYPE_ID_UINT64, &sleep_max);
+	return stress_set_setting("sleep-max", TYPE_ID_UINT64, &sleep_max);
 }
 
-void stress_adjust_sleep_max(const uint64_t max)
+static const stress_opt_set_func_t opt_set_funcs[] = {
+	{ OPT_sleep_max,	stress_set_sleep_max },
+	{ 0,			NULL }
+};
+
+#if defined(HAVE_LIB_PTHREAD)
+
+static void MLOCKED_TEXT stress_sigalrm_handler(int signum)
 {
-	uint64_t sleep_max_adjustment = max;
+	(void)signum;
 
-	set_setting("sleep-max-adjustment", TYPE_ID_UINT64,
-		&sleep_max_adjustment);
+	thread_terminate = true;
 }
 
-#if defined(HAVE_LIB_PTHREAD) && defined(__linux__)
-
 /*
  *  stress_pthread_func()
  *	pthread that performs different ranges of sleeps
  */
 static void *stress_pthread_func(void *c)
 {
-	uint8_t stack[SIGSTKSZ + STACK_ALIGNMENT];
 	static void *nowt = NULL;
-	ctxt_t *ctxt = (ctxt_t *)c;
+	stress_ctxt_t *ctxt = (stress_ctxt_t *)c;
+	const stress_args_t *args = ctxt->args;
+	const uint64_t max_ops =
+		args->max_ops ? (args->max_ops / ctxt->sleep_max) + 1 : 0;
 
-	/*
-	 *  Block all signals, let controlling thread
-	 *  handle these
-	 */
-	(void)sigprocmask(SIG_BLOCK, &set, NULL);
-
-	/*
-	 *  According to POSIX.1 a thread should have
-	 *  a distinct alternative signal stack.
-	 *  However, we block signals in this thread
-	 *  so this is probably just totally unncessary.
-	 */
-	(void)memset(stack, 0, sizeof(stack));
-	if (stress_sigaltstack(stack, SIGSTKSZ) < 0)
-		goto die;
-
-	while (!thread_terminate) {
+	while (keep_stressing(args) &&
+	       !thread_terminate &&
+	       (!max_ops || (ctxt->counter < max_ops))) {
 		struct timespec tv;
+#if defined(HAVE_SYS_SELECT_H)
 		struct timeval timeout;
+#endif
 
 		tv.tv_sec = 0;
 		tv.tv_nsec = 1;
@@ -110,6 +110,8 @@ static void *stress_pthread_func(void *c)
 			break;
 		if (shim_usleep(10000) < 0)
 			break;
+
+#if defined(HAVE_SYS_SELECT_H)
 		timeout.tv_sec = 0;
 		timeout.tv_usec = 10;
 		if (select(0, NULL, NULL, NULL, &timeout) < 0)
@@ -126,10 +128,10 @@ static void *stress_pthread_func(void *c)
 		timeout.tv_usec = 10000;
 		if (select(0, NULL, NULL, NULL, &timeout) < 0)
 			break;
+#endif
 
 		ctxt->counter++;
 	}
-die:
 	return &nowt;
 }
 
@@ -137,34 +139,31 @@ die:
  *  stress_sleep()
  *	stress by many sleeping threads
  */
-int stress_sleep(const args_t *args)
+static int stress_sleep(const stress_args_t *args)
 {
 	uint64_t i, n, limited = 0;
-	uint64_t sleep_max = DEFAULT_SLEEP, sleep_max_adjustment;
-	ctxt_t    ctxts[MAX_SLEEP];
+	uint64_t sleep_max = DEFAULT_SLEEP;
+	static stress_ctxt_t ctxts[MAX_SLEEP];
 	int ret = EXIT_SUCCESS;
-	bool ok = true;
 
-	if (!get_setting("sleep-max", &sleep_max)) {
+	if (!stress_get_setting("sleep-max", &sleep_max)) {
 		if (g_opt_flags & OPT_FLAGS_MAXIMIZE)
 			sleep_max = MAX_SLEEP;
 		if (g_opt_flags & OPT_FLAGS_MINIMIZE)
 			sleep_max = MIN_SLEEP;
 	}
 
-	if (get_setting("sleep-max-adjustment", &sleep_max_adjustment)) {
-		if (sleep_max > sleep_max_adjustment) {
-			sleep_max = sleep_max_adjustment;
-			pr_inf("re-adjusting maximum threads to "
-				"soft limit f %" PRIu64 "\n",
-				sleep_max);
-		}
-	}
+	if (stress_sighandler(args->name, SIGALRM, stress_sigalrm_handler, NULL) < 0)
+		return EXIT_FAILURE;
+
 	(void)memset(ctxts, 0, sizeof(ctxts));
 	(void)sigfillset(&set);
 
-	for (n = 0; n < sleep_max;  n++) {
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
+
+	for (n = 0; n < sleep_max; n++) {
 		ctxts[n].args = args;
+		ctxts[n].sleep_max = sleep_max;
 		ret = pthread_create(&ctxts[n].pthread, NULL,
 			stress_pthread_func, &ctxts[n]);
 		if (ret) {
@@ -174,30 +173,35 @@ int stress_sleep(const args_t *args)
 				break;
 			}
 			/* Something really unexpected */
-			pr_fail_errno("pthread create", ret);
+			pr_fail("%s: pthread create failed, errno=%d (%s)\n",
+				args->name, ret, strerror(ret));
 			ret = EXIT_NO_RESOURCE;
 			goto tidy;
 		}
 		/* Timed out? abort! */
-		if (!g_keep_stressing_flag)
+		if (!keep_stressing_flag())
 			goto tidy;
 	}
 
 	do {
-		*args->counter = 0;
-		(void)shim_usleep(10000);
+		set_counter(args, 0);
+		(void)shim_usleep_interruptible(10000);
 		for (i = 0; i < n; i++)
-			*args->counter += ctxts[i].counter;
-	}  while (ok && keep_stressing());
-
+			add_counter(args, ctxts[i].counter);
+	}  while (!thread_terminate && keep_stressing(args));
 
 	ret = EXIT_SUCCESS;
 tidy:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
+
 	thread_terminate = true;
 	for (i = 0; i < n; i++) {
 		ret = pthread_join(ctxts[i].pthread, NULL);
+		(void)ret;
+		/*
 		if (ret)
-			pr_dbg("pthread join, ret=%d\b", ret);
+			pr_dbg("%s: pthread join, ret=%d\n", args->name, ret);
+		*/
 	}
 
 	if (limited) {
@@ -211,9 +215,18 @@ tidy:
 
 	return ret;
 }
+
+stressor_info_t stress_sleep_info = {
+	.stressor = stress_sleep,
+	.class = CLASS_INTERRUPT | CLASS_SCHEDULER | CLASS_OS,
+	.opt_set_funcs = opt_set_funcs,
+	.help = help
+};
 #else
-int stress_sleep(const args_t *args)
-{
-	return stress_not_implemented(args);
-}
+stressor_info_t stress_sleep_info = {
+	.stressor = stress_not_implemented,
+	.class = CLASS_INTERRUPT | CLASS_SCHEDULER | CLASS_OS,
+	.opt_set_funcs = opt_set_funcs,
+	.help = help
+};
 #endif
