@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2013-2017 Canonical, Ltd.
+ * Copyright (C) 2013-2021 Canonical, Ltd.
  *
  * This program is free software; you can redistribute it and/or
  * modify it under the terms of the GNU General Public License
@@ -24,58 +24,74 @@
  */
 #include "stress-ng.h"
 
-#if defined(__linux__) && defined(__NR_userfaultfd)
+static const stress_help_t help[] = {
+	{ NULL,	"userfaultfd N",	"start N page faulting workers with userspace handling" },
+	{ NULL,	"userfaultfd-ops N",	"stop after N page faults have been handled" },
+	{ NULL,	NULL,			NULL }
+};
 
-#include <poll.h>
-#include <linux/userfaultfd.h>
+#if defined(HAVE_USERFAULTFD) && 	 \
+    defined(HAVE_LINUX_USERFAULTFD_H) && \
+    defined(HAVE_POLL_H) &&		 \
+    defined(HAVE_CLONE)
 
 #define STACK_SIZE	(64 * 1024)
+#define COUNT_MAX	(256)
 
 /* Context for clone */
 typedef struct {
-	const args_t *args;
+	const stress_args_t *args;
 	uint8_t *data;
 	size_t page_size;
 	size_t sz;
 	pid_t parent;
-} context_t;
+} stress_context_t;
 
 #endif
 
-void stress_set_userfaultfd_bytes(const char *opt)
+static int stress_set_userfaultfd_bytes(const char *opt)
 {
 	size_t userfaultfd_bytes;
 
-	userfaultfd_bytes = (size_t)get_uint64_byte_memory(opt, 1);
-	check_range_bytes("userfaultfd-bytes", userfaultfd_bytes,
+	userfaultfd_bytes = (size_t)stress_get_uint64_byte_memory(opt, 1);
+	stress_check_range_bytes("userfaultfd-bytes", userfaultfd_bytes,
 		MIN_MMAP_BYTES, MAX_MEM_LIMIT);
-	set_setting("userfaultfd-bytes", TYPE_ID_SIZE_T, &userfaultfd_bytes);
+	return stress_set_setting("userfaultfd-bytes", TYPE_ID_SIZE_T, &userfaultfd_bytes);
 }
 
-#if defined(__linux__) && defined(__NR_userfaultfd)
+static const stress_opt_set_func_t opt_set_funcs[] = {
+	{ OPT_userfaultfd_bytes,	stress_set_userfaultfd_bytes },
+	{ 0,				NULL }
+};
+
+#if defined(HAVE_USERFAULTFD) && \
+    defined(HAVE_LINUX_USERFAULTFD_H) && \
+    defined(HAVE_CLONE)
 
 /*
  *  stress_child_alarm_handler()
  *	SIGALRM handler to terminate child immediately
  */
-static void MLOCKED stress_child_alarm_handler(int dummy)
+static void MLOCKED_TEXT NORETURN stress_child_alarm_handler(int signum)
 {
-	(void)dummy;
+	(void)signum;
 
 	_exit(0);
 }
 
 /*
- *  stress_userfaultfd_child()
+ *  stress_userfaultfd_clone()
  *	generate page faults for parent to handle
  */
-static int stress_userfaultfd_child(void *arg)
+static int stress_userfaultfd_clone(void *arg)
 {
-	context_t *c = (context_t *)arg;
-	const args_t *args = c->args;
+	stress_context_t *c = (stress_context_t *)arg;
+	const stress_args_t *args = c->args;
 
 	(void)setpgid(0, g_pgrp);
 	stress_parent_died_alarm();
+	(void)sched_settings_apply(true);
+
 	if (stress_sighandler(args->name, SIGALRM, stress_child_alarm_handler, NULL) < 0)
 		return EXIT_NO_RESOURCE;
 
@@ -84,14 +100,15 @@ static int stress_userfaultfd_child(void *arg)
 
 		/* hint we don't need these pages */
 		if (shim_madvise(c->data, c->sz, MADV_DONTNEED) < 0) {
-			pr_fail_err("userfaultfd madvise failed");
+			pr_fail("%s: madvise failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			(void)kill(c->parent, SIGALRM);
 			return -1;
 		}
 		/* and trigger some page faults */
 		for (ptr = c->data; ptr < end; ptr += c->page_size)
 			*ptr = 0xff;
-	} while (keep_stressing());
+	} while (keep_stressing(args));
 
 	return 0;
 }
@@ -101,20 +118,20 @@ static int stress_userfaultfd_child(void *arg)
  *	handle a write page fault caused by child
  */
 static inline int handle_page_fault(
-	const args_t *args,
-	int fd,
+	const stress_args_t *args,
+	const int fd,
 	uint8_t *addr,
 	void *zero_page,
-	uint8_t *data_start,
-	uint8_t *data_end,
-	size_t page_size)
+	const uint8_t *data_start,
+	const uint8_t *data_end,
+	const size_t page_size)
 {
 	if ((addr < data_start) || (addr >= data_end)) {
-		pr_fail_err("userfaultfd page fault address out of range");
+		pr_fail("%s: page fault address is out of range\n", args->name);
 		return -1;
 	}
 
-	if (mwc32() & 1) {
+	if (stress_mwc32() & 1) {
 		struct uffdio_copy copy;
 
 		copy.copy = 0;
@@ -124,7 +141,8 @@ static inline int handle_page_fault(
 		copy.len = page_size;
 
 		if (ioctl(fd, UFFDIO_COPY, &copy) < 0) {
-			pr_fail_err("userfaultfd page fault copy ioctl failed");
+			pr_fail("%s: page fault ioctl UFFDIO_COPY failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			return -1;
 		}
 	} else {
@@ -134,7 +152,8 @@ static inline int handle_page_fault(
 		zeropage.range.len = page_size;
 		zeropage.mode = 0;
 		if (ioctl(fd, UFFDIO_ZEROPAGE, &zeropage) < 0) {
-			pr_fail_err("userfaultfd page fault zeropage ioctl failed");
+			pr_fail("%s: page fault ioctl UFFDIO_ZEROPAGE failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
 			return -1;
 		}
 	}
@@ -147,28 +166,38 @@ static inline int handle_page_fault(
  *	is an OOM-able child process that the
  *	parent can restart
  */
-static int stress_userfaultfd_oomable(
-	const args_t *args,
-	size_t userfaultfd_bytes)
+static int stress_userfaultfd_child(const stress_args_t *args, void *context)
 {
 	const size_t page_size = args->page_size;
 	size_t sz;
 	uint8_t *data;
 	void *zero_page = NULL;
-	int fd = -1, status, rc = EXIT_SUCCESS;
+	int fd = -1, status, rc = EXIT_SUCCESS, count = 0;
 	const unsigned int uffdio_copy = 1 << _UFFDIO_COPY;
 	const unsigned int uffdio_zeropage = 1 << _UFFDIO_ZEROPAGE;
 	pid_t pid;
+	const pid_t self = getpid();
 	struct uffdio_api api;
 	struct uffdio_register reg;
-	context_t c;
+	stress_context_t c;
 	bool do_poll = true;
+	static uint8_t stack[STACK_SIZE]; /* Child clone stack */
+	uint8_t *stack_top = (uint8_t *)stress_get_stack_top((void *)stack, STACK_SIZE);
+	size_t userfaultfd_bytes = DEFAULT_MMAP_BYTES;
 
-	/* Child clone stack */
-	static uint8_t stack[STACK_SIZE];
-	const ssize_t stack_offset =
-		stress_get_stack_direction() * (STACK_SIZE - 64);
-	uint8_t *stack_top = stack + stack_offset;
+	(void)context;
+
+	if (!stress_get_setting("userfaultfd-bytes", &userfaultfd_bytes)) {
+		if (g_opt_flags & OPT_FLAGS_MAXIMIZE)
+			userfaultfd_bytes = MAX_32;
+		if (g_opt_flags & OPT_FLAGS_MINIMIZE)
+			userfaultfd_bytes = MIN_MMAP_BYTES;
+	}
+	userfaultfd_bytes /= args->num_instances;
+	if (userfaultfd_bytes < MIN_MMAP_BYTES)
+		userfaultfd_bytes = MIN_MMAP_BYTES;
+	if (userfaultfd_bytes < args->page_size)
+		userfaultfd_bytes = args->page_size;
 
 	sz = userfaultfd_bytes & ~(page_size - 1);
 
@@ -185,11 +214,34 @@ static int stress_userfaultfd_oomable(
 		goto free_zeropage;
 	}
 
+	/* Exercise invalid flags */
+	fd = shim_userfaultfd(~0);
+	if (fd >= 0)
+		(void)close(fd);
+
 	/* Get userfault fd */
-	if ((fd = shim_userfaultfd(0)) < 0) {
-		rc = exit_status(errno);
-		pr_err("%s: userfaultfd failed, errno = %d (%s)\n",
-			args->name, errno, strerror(errno));
+	fd = shim_userfaultfd(0);
+	if (fd < 0) {
+		switch (errno) {
+		case EPERM:
+			if (args->instance == 0)
+				pr_inf_skip("%s: stressor will be skipped, insufficient "
+					"privilege\n", args->name);
+			rc = EXIT_NO_RESOURCE;
+			break;
+		case ENOSYS:
+			if (args->instance == 0)
+				pr_inf_skip("%s: stressor will be skipped, "
+					"userfaultfd not supported\n",
+					args->name);
+			rc = EXIT_NOT_IMPLEMENTED;
+			break;
+		default:
+			rc = exit_status(errno);
+			pr_err("%s: userfaultfd failed, errno = %d (%s)\n",
+				args->name, errno, strerror(errno));
+			break;
+		}
 		goto unmap_data;
 	}
 
@@ -245,13 +297,15 @@ static int stress_userfaultfd_oomable(
 	c.data = data;
 	c.sz = sz;
 	c.page_size = page_size;
-	c.parent = getpid();
+	c.parent = self;
+
+	stress_set_proc_state(args->name, STRESS_STATE_RUN);
 
 	/*
-	 *  We need to clone the child and share the same VM address space
+	 *  We need to clone and share the same VM address space
 	 *  as parent so we can perform the page fault handling
 	 */
-	pid = clone(stress_userfaultfd_child, align_stack(stack_top),
+	pid = clone(stress_userfaultfd_clone, stress_align_stack(stack_top),
 		SIGCHLD | CLONE_FILES | CLONE_FS | CLONE_SIGHAND | CLONE_VM, &c);
 	if (pid < 0) {
 		pr_err("%s: fork failed, errno = %d (%s)\n",
@@ -262,10 +316,11 @@ static int stress_userfaultfd_oomable(
 	/* Parent */
 	do {
 		struct uffd_msg msg;
+		struct uffdio_range wake;
 		ssize_t ret;
 
 		/* check we should break out before we block on the read */
-		if (!g_keep_stressing_flag)
+		if (!keep_stressing_flag())
 			break;
 
 		/*
@@ -287,8 +342,9 @@ static int stress_userfaultfd_oomable(
 				if (errno == EINTR)
 					continue;
 				if (errno != ENOMEM) {
-					pr_fail_err("poll userfaultfd");
-					if (!g_keep_stressing_flag)
+					pr_fail("%s: poll failed, errno=%d (%s)\n",
+						args->name, errno, strerror(errno));
+					if (!keep_stressing_flag())
 						break;
 				}
 				/*
@@ -301,41 +357,57 @@ static int stress_userfaultfd_oomable(
 			/* No data, re-poll */
 			if (!(fds[0].revents & POLLIN))
 				continue;
+
+			if (UNLIKELY(count++ >= COUNT_MAX)) {
+				(void)stress_read_fdinfo(self, fd);
+				count = 0;
+			}
 		}
 
 do_read:
-		if ((ret = read(fd, &msg, sizeof(msg))) < 0) {
+		ret = read(fd, &msg, sizeof(msg));
+		if (ret < 0) {
 			if (errno == EINTR)
 				continue;
-			pr_fail_err("read userfaultfd");
-			if (!g_keep_stressing_flag)
+			pr_fail("%s: read failed, errno=%d (%s)\n",
+				args->name, errno, strerror(errno));
+			if (!keep_stressing_flag())
 				break;
 			continue;
 		}
 		/* We only expect a page fault event */
 		if (msg.event != UFFD_EVENT_PAGEFAULT) {
-			pr_fail_err("userfaultfd msg not pagefault event");
+			pr_fail("%s: msg event not a pagefault event\n", args->name);
 			continue;
 		}
 		/* We only expect a write fault */
 		if (!(msg.arg.pagefault.flags & UFFD_PAGEFAULT_FLAG_WRITE)) {
-			pr_fail_err("userfaultfd msg not write page fault event");
+			pr_fail("%s: msg event not write page fault event\n", args->name);
 			continue;
 		}
 		/* Go handle the page fault */
-		if (handle_page_fault(args, fd, (uint8_t *)(ptrdiff_t)msg.arg.pagefault.address,
+		if (handle_page_fault(args, fd, (uint8_t *)(intptr_t)msg.arg.pagefault.address,
 				zero_page, data, data + sz, page_size) < 0)
 			break;
-		inc_counter(args);
-	} while (keep_stressing());
 
+		(void)memset(&wake, 0, sizeof(wake));
+		wake.start = (uintptr_t)data;
+		wake.len = page_size;
+		ret = ioctl(fd, UFFDIO_WAKE, &wake);
+		(void)ret;
+
+		inc_counter(args);
+	} while (keep_stressing(args));
+
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 	/* Run it over, zap child */
 	(void)kill(pid, SIGKILL);
-	if (waitpid(pid, &status, 0) < 0) {
+	if (shim_waitpid(pid, &status, 0) < 0) {
 		pr_dbg("%s: waitpid failed, errno = %d (%s)\n",
 			args->name, errno, strerror(errno));
 	}
 unreg:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 	if (ioctl(fd, UFFDIO_UNREGISTER, &reg) < 0) {
 		pr_err("%s: ioctl UFFDIO_UNREGISTER failed, errno = %d (%s)\n",
 			args->name, errno, strerror(errno));
@@ -343,9 +415,14 @@ unreg:
 		goto unmap_data;
 	}
 unmap_data:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 	(void)munmap(data, sz);
 free_zeropage:
+	stress_set_proc_state(args->name, STRESS_STATE_DEINIT);
 	free(zero_page);
+	if (fd > -1)
+		(void)close(fd);
+
 	return rc;
 }
 
@@ -353,71 +430,22 @@ free_zeropage:
  *  stress_userfaultfd()
  *	stress userfaultfd
  */
-int stress_userfaultfd(const args_t *args)
+static int stress_userfaultfd(const stress_args_t *args)
 {
-	pid_t pid;
-	int rc = EXIT_FAILURE;
-	size_t userfaultfd_bytes = DEFAULT_MMAP_BYTES;
-
-	if (!get_setting("userfaultfd-bytes", &userfaultfd_bytes)) {
-		if (g_opt_flags & OPT_FLAGS_MAXIMIZE)
-			userfaultfd_bytes = MAX_MMAP_BYTES;
-		if (g_opt_flags & OPT_FLAGS_MINIMIZE)
-			userfaultfd_bytes = MIN_MMAP_BYTES;
-	}
-	userfaultfd_bytes /= args->num_instances;
-	if (userfaultfd_bytes < MIN_MMAP_BYTES)
-		userfaultfd_bytes = MIN_MMAP_BYTES;
-	if (userfaultfd_bytes < args->page_size)
-		userfaultfd_bytes = args->page_size;
-
-	pid = fork();
-	if (pid < 0) {
-		if (errno == EAGAIN)
-			return EXIT_NO_RESOURCE;
-		pr_err("%s: fork failed: errno=%d: (%s)\n",
-			args->name, errno, strerror(errno));
-	} else if (pid > 0) {
-		/* Parent */
-		int status, ret;
-
-		(void)setpgid(pid, g_pgrp);
-		ret = waitpid(pid, &status, 0);
-		if (ret < 0) {
-			if (errno != EINTR)
-				pr_dbg("%s: waitpid(): errno=%d (%s)\n",
-					args->name, errno, strerror(errno));
-			(void)kill(pid, SIGTERM);
-			(void)kill(pid, SIGKILL);
-			(void)waitpid(pid, &status, 0);
-		} else if (WIFSIGNALED(status)) {
-			pr_dbg("%s: child died: %s (instance %d)\n",
-				args->name, stress_strsignal(WTERMSIG(status)),
-				args->instance);
-			/* If we got killed by OOM killer, report this */
-			if (WTERMSIG(status) == SIGKILL) {
-				log_system_mem_info();
-				pr_dbg("%s: assuming killed by OOM "
-					"killer, aborting "
-					"(instance %d)\n",
-					args->name, args->instance);
-				return EXIT_NO_RESOURCE;
-			}
-			return EXIT_FAILURE;
-		}
-		rc = WEXITSTATUS(status);
-	} else if (pid == 0) {
-		/* Child */
-		(void)setpgid(0, g_pgrp);
-		stress_parent_died_alarm();
-
-		_exit(stress_userfaultfd_oomable(args, userfaultfd_bytes));
-	}
-	return rc;
+	return stress_oomable_child(args, NULL, stress_userfaultfd_child, STRESS_OOMABLE_NORMAL);
 }
+
+stressor_info_t stress_userfaultfd_info = {
+	.stressor = stress_userfaultfd,
+	.class = CLASS_VM | CLASS_OS,
+	.opt_set_funcs = opt_set_funcs,
+	.help = help
+};
 #else
-int stress_userfaultfd(const args_t *args)
-{
-	return stress_not_implemented(args);
-}
+stressor_info_t stress_userfaultfd_info = {
+	.stressor = stress_not_implemented,
+	.class = CLASS_VM | CLASS_OS,
+	.opt_set_funcs = opt_set_funcs,
+	.help = help
+};
 #endif
